#include "RectUtils.h"

#include "AdvancedOptionsView.h"
#include "BitmapMovie.h"
#include "BSCWindow.h"
#include "messages.h"
#include "OutputView.h"
#include "SelectionWindow.h"

#include <Application.h>
#include <Bitmap.h>
#include <Box.h>
#include <Button.h>
#include <Entry.h>
#include <MediaFile.h>
#include <MediaTrack.h>
#include <Screen.h>
#include <StatusBar.h>
#include <String.h>
#include <StringView.h>
#include <TabView.h>
#include <TextControl.h>

#include <cstdio>
#include <cstdlib>

const BRect kWindowRect(0, 0, 480, 300);

BSCWindow::BSCWindow()
	:
	BDirectWindow(kWindowRect, "BeScreenCapture", B_TITLED_WINDOW,
		B_ASYNCHRONOUS_CONTROLS|B_NOT_RESIZABLE|B_NOT_ZOOMABLE),
	fCaptureThread(-1),
	fSaverThread(-1)
{
	center_window_on_screen(this);
	
	BBox *boundingBox = new BBox(Bounds(), "boundingBox");
	boundingBox->ResizeBy(1, 1);
	
	BRect tabRect = boundingBox->Bounds();
	tabRect.InsetBy(10, 10);
	tabRect.bottom -= 40;
	fTabView = new BTabView(tabRect, "Tab View");
	TOutputView *outputView = new TOutputView(fTabView->ContainerView()->Bounds());
	TAdvancedOptionsView *advancedView = new TAdvancedOptionsView(fTabView->ContainerView()->Bounds());
	fTabView->AddTab(outputView);
	fTabView->AddTab(advancedView);
	
	boundingBox->AddChild(fTabView);
	
	BRect buttonRect = boundingBox->Frame();
	buttonRect.bottom -= 10;
	buttonRect.top = buttonRect.bottom - 30;
	buttonRect.right -= 45;
	buttonRect.left = buttonRect.right - 60;
	fStartStopButton = new BButton(buttonRect, "Start", "Start Recording", new BMessage(kToggleCapture)); 
	boundingBox->AddChild(fStartStopButton);
	fStartStopButton->ResizeToPreferred();
	
	const char *kString = "Finishing writing the movie to disk...";
	BRect stringRect;
	stringRect.left = 15;
	stringRect.top = boundingBox->Frame().bottom - 35;
	stringRect.right = 100;
	stringRect.bottom = stringRect.top + 20;
	
	fStringView = new BStringView(stringRect, "stringview", kString);
	boundingBox->AddChild(fStringView);
	fStringView->ResizeToPreferred();
	fStringView->Hide();	

	BRect statusRect = stringRect;
	statusRect.left = stringRect.left + fStringView->StringWidth(kString) + 10;
	statusRect.right = statusRect.left + 60;
	statusRect.bottom -= 15;
	statusRect.top -= 15;
	fStatusBar = new BStatusBar(statusRect, "", "");
	boundingBox->AddChild(fStatusBar);
	fStatusBar->Hide();
		
	AddChild(boundingBox);
}


bool
BSCWindow::QuitRequested()
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	
	return (fSaverThread < 0) && (fCaptureThread < 0) && BWindow::QuitRequested();
}


void
BSCWindow::MessageReceived(BMessage *message)
{
	switch (message->what) {
		case kToggleCapture:
			if (fCaptureThread < 0)
				StartCapture();
			else
				EndCapture();
			break;
		case kSelectArea:
		{
			Minimize(true);
			TSelectionWindow *window = new TSelectionWindow();
			window->SetTarget(this);
			window->SetCommand(kAreaSelected);
			window->Show();
			break;
		}
		case kAreaSelected:
		{
			BRect rect;
			if (message->FindRect("selection", &rect) == B_OK)
				static_cast<TOutputView *>(fTabView->ViewForTab(0))->SetCaptureArea(rect);
			Minimize(false);
			break;
		}
		case kSaverReturns:
		{
			status_t unused;
			wait_for_thread(fSaverThread, &unused);
			fSaverThread = -1;
			fStartStopButton->SetEnabled(true);
			fStringView->Hide();
			fStatusBar->Hide();
			delete fBitmaps;
			delete fCursorPositions;
			break;
		}
		case kFPSChanged:
		{
			TOutputView *view = static_cast<TOutputView *>(fTabView->ViewForTab(0));
			void *ptr;
			if (message->FindPointer("source", &ptr) == B_OK) {
				BTextControl *control = static_cast<BTextControl *>(ptr);
				view->SetPlayingFPS(atoi(control->Text()));
				view->RebuildCodecMenu();
			}
			break;
		}
		default:
			BWindow::MessageReceived(message);
			break;
	}
}


void
BSCWindow::ScreenChanged(BRect screen_size, color_space depth)
{
	TOutputView *view = static_cast<TOutputView *>(fTabView->ViewForTab(0));
	if (view != NULL) {
		view->RebuildCodecMenu();
	}
}


void
BSCWindow::DirectConnected(direct_buffer_info *info)
{
	//if (fDirectLock.Lock()) {
		switch (info->buffer_state & B_DIRECT_MODE_MASK) {
			case B_DIRECT_START:
			case B_DIRECT_MODIFY:
				fDirectInfo = *info;
				break;
			case B_DIRECT_STOP:
				break;
			default:
				break;
		}
	//	fDirectLock.Unlock();
	//}
}


status_t
BSCWindow::GetBitmap(BBitmap **bitmap, BRect bounds)
{
	BBitmap *newBitmap = new BBitmap(bounds, fDirectInfo.pixel_format);
	if (newBitmap == NULL)
		return B_ERROR;
	
	*bitmap = newBitmap;
	return ReadBitmap(newBitmap, bounds);
}


status_t
BSCWindow::ReadBitmap(BBitmap *bitmap, BRect bounds)
{
	int32 bytesPerPixel = fDirectInfo.bits_per_pixel >> 3;
	int32 size = (bounds.IntegerWidth() + 1) * bytesPerPixel;
	uint32 rowBytes = fDirectInfo.bytes_per_row / bytesPerPixel;	    		
	int32 offset = ((uint32)bounds.left + ((uint32)bounds.top * rowBytes)) * bytesPerPixel;  				
	
	int32 bitmapBytesPerRow = bitmap->BytesPerRow();
	int32 height = bounds.IntegerHeight() + 1;
	
	//if (fDirectLock.Lock()) {
		void *from = (void *)((uint8 *)fDirectInfo.bits + offset);    		
	    void *to = bitmap->Bits();
	    		
	    for (int32 y = 0; y < height; y++) {  
	    	memcpy(to, from, size);	
	    	to = (void *)((uint8 *)to + bitmapBytesPerRow);
	    	from = (void *)((uint8 *)from + fDirectInfo.bytes_per_row);
	    }
	//    fDirectLock.Unlock();
	    
	    return B_OK;
	//}
	
	return B_ERROR;
}


void
BSCWindow::StartCapture()
{
	printf("StartCapture\n");
	TOutputView *outputView = static_cast<TOutputView *>(fTabView->ViewForTab(0));
	TAdvancedOptionsView *advOpts = static_cast<TAdvancedOptionsView *>(fTabView->ViewForTab(1));			
		
	// Initializing capture options
	fBitmapCount = 0;	
	memset(&fCaptureOptions, 0, sizeof(fCaptureOptions));
	fCaptureOptions.fileName = outputView->OutputFileName();
	fCaptureOptions.includeCursor = outputView->IncludeCursor();
	fCaptureOptions.framesPerSecond = advOpts->FramesPerSecond();
	printf("frames per sec: %ld\n", fCaptureOptions.framesPerSecond);
	fCaptureOptions.bounds = outputView->CaptureArea();
	fCaptureOptions.formatFamily = outputView->FormatFamily();
	fCaptureOptions.format = outputView->Format();
	fCaptureOptions.codecInfo = outputView->Codec();
	fCaptureOptions.useDirectWindow = advOpts->UseDirectWindow();
	fCaptureOptions.notifyStatusBar = false;
	fStatusBar->Reset();
	
	if (outputView->MinimizeOnStart())
		Minimize(true);
	
	fBitmaps = new queue <BBitmap *>;
	fCursorPositions = new queue <BPoint >;
	
	fCaptureThread = spawn_thread((thread_entry)CaptureThread, "Capture Thread",
					B_DISPLAY_PRIORITY, this);
	fSaverThread = spawn_thread((thread_entry)SaverThread, "Saver Thread",
					B_LOW_PRIORITY, this);
	
	if (fCaptureThread > 0 && fSaverThread > 0) {
		fStartStopButton->SetLabel("Stop Recording");
		//if (fDirectLock.IsLocked())
		//	fDirectLock.Unlock();
		fKillThread = false;
		resume_thread(fCaptureThread);
		resume_thread(fSaverThread);
	} else {
		kill_thread(fCaptureThread);
		kill_thread(fSaverThread);
		fCaptureThread = -1;
		fSaverThread = -1;
	}
}


void
BSCWindow::EndCapture()
{
	printf("EndCapture\n");
	
	fKillThread = true;
	
	//if (fDirectLock.IsLocked())
	//	fDirectLock.Unlock();
		
	if (fCaptureThread > 0) {
		status_t unused;
		wait_for_thread(fCaptureThread, &unused);
		fStartStopButton->SetEnabled(false);
		fStartStopButton->SetLabel("Start Recording");
		fCaptureThread = -1;
		
	}
	
	if (fSaverThread > 0) {
		set_thread_priority(fSaverThread, B_NORMAL_PRIORITY);
		fStringView->Show();
		fStatusBar->Show();
		fCaptureOptions.notifyStatusBar = true;
	}
}


int
BSCWindow::CaptureThread(void *arg)
{
	BSCWindow *object = static_cast<BSCWindow *>(arg);
	bool useDirectWindow = object->fCaptureOptions.useDirectWindow;
	
	// To get mouse positions
	BView *dummyView = NULL;
	if (object->LockLooper()) {
		dummyView = object->FindView("dummy");
		if (dummyView == NULL) {
			dummyView = new BView(BRect(0, 0, 0, 0), "dummy", B_FOLLOW_NONE, 0);
			object->AddChild(dummyView);
		}
		object->UnlockLooper();
	}
	
	BRect bounds = object->fCaptureOptions.bounds;
	BScreen screen;
	
	bigtime_t lastFrameTime = 0;
	bigtime_t waitTime = 1000000 / object->fCaptureOptions.framesPerSecond;
	
	BPoint mousePos;
	ulong buttons;
	
	while (!object->fKillThread) {
		lastFrameTime = system_time();			
		status_t error = B_ERROR;
		BBitmap *bitmap = NULL;
		
		if (useDirectWindow)		
			error = object->GetBitmap(&bitmap, bounds);			
		else 
			error = screen.GetBitmap(&bitmap, false, &bounds);
				 
	    if (error == B_OK) {
	    	/*if (object->LockLooper()) {
	    		printf("Looper locked\n");
				dummyView->GetMouse(&mousePos, &buttons, false);
				dummyView->ConvertToScreen(&mousePos);
				printf("MouseX: %d\n", (int32)mousePos.x);
				object->UnlockLooper();
				printf("Looper released\n");
			}
			*/
			if (object->fLocker.Lock()) {
				object->fBitmaps->push(bitmap);			
				//object->fCursorPositions->push(mousePos);
				object->fBitmapCount++;
				object->fLocker.Unlock();
			}
		} else {
			printf("Error while getting screenshot: %s\n", strerror(error));
			return -1;
		}
		
		snooze_until(lastFrameTime + waitTime, B_SYSTEM_TIMEBASE);
		screen.WaitForRetrace();
	}
	
	return 0;
}


int
BSCWindow::SaverThread(void *arg)
{
	BSCWindow *object = static_cast<BSCWindow *>(arg);
	
	BScreen *screen = new BScreen(object);
	BRect bounds = object->fCaptureOptions.bounds;
	int32 newWidth = 300;
	int32 newHeight = 200;
	BBitmap *resizedBitmap = new BBitmap(BRect(0, 0, newWidth - 1, newHeight - 1), screen->ColorSpace());
	BitmapMovie *bitmapMovie = new BitmapMovie((uint32)bounds.Width() + 1,
			(uint32)bounds.Height() + 1, screen->ColorSpace());

	delete screen;
	
	bool okToWrite = true;
	
	// fill in the list of available file formats	
	int32 appendInt = 0;
	entry_ref ref;
	get_ref_for_path(object->fCaptureOptions.fileName.Path(), &ref);
	while (BEntry(&ref).Exists()) {
		char tmp[32];
		appendInt++;
		BPath path(object->fCaptureOptions.fileName.Path());
		snprintf(tmp, 32, " %ld", appendInt);
		BString str;
		str << path.Leaf() << tmp;
		path.GetParent(&path);
		path.Append(str.String());
		get_ref_for_path(path.Path(), &ref);
	}
	
	media_file_format mff;
	int32 cookie = 0;
	while (get_next_file_format(&cookie, &mff) == B_OK) {
		if (mff.family == object->fCaptureOptions.formatFamily)
			break;
	}
		
	media_format outFormat = object->fCaptureOptions.format; 
	media_codec_info codec = object->fCaptureOptions.codecInfo;
	
	printf("Creating file...%s\n", ref.name);
	status_t status = bitmapMovie->CreateFile(ref, mff, outFormat, codec);
	if (status != B_OK) {
		printf("Couldn't create movie file: %s\n", strerror(status));
		okToWrite = false;
	}
	
	BMessage progressMessage(B_UPDATE_STATUS_BAR);
	
	int32 framesWritten = 0;
	int32 framesLeft = 0; // filled later, used just for the BStatusBar
	bool firstFrame = true;
	BBitmap *bitmap = NULL;
	BPoint mousePt;
	while (!object->fKillThread || object->fBitmapCount > 0) {
		if (object->fBitmapCount > 0) {				
			if (object->fLocker.Lock()) {
				bitmap = object->fBitmaps->front();
				object->fBitmaps->pop();
				//mousePt = object->fCursorPositions->front();			
				//object->fCursorPositions->pop();
				object->fLocker.Unlock();
			} else
				okToWrite = false;
			
			if (okToWrite) {
				status_t error = B_OK;
				error = bitmapMovie->WriteFrame(bitmap, firstFrame);
				if (error == B_OK) {
					framesWritten++;
					firstFrame = false;				
				} else {
					printf("%s\n", strerror(error));
					break;
				}
			}
			
			delete bitmap;
			atomic_add(&object->fBitmapCount, -1);
			
			if (object->fCaptureOptions.notifyStatusBar) {
				if (progressMessage.HasFloat("delta"))
					progressMessage.RemoveData("delta");
				else
					framesLeft = object->fBitmapCount;
				
				progressMessage.AddFloat("delta", (1 / (float)framesLeft) * 100);
				object->PostMessage(&progressMessage, object->fStatusBar);
			}	
		}
	}
	
	bitmapMovie->CloseFile();
	delete bitmapMovie;
	
	delete resizedBitmap;
	printf("%ld frames written\n", framesWritten);
	
	object->PostMessage(kSaverReturns);
	
	return 0;
}
