#include "BitmapMovie.h"
#include "OutputView.h"
#include "RectView.h"
#include "messages.h"

#include <Bitmap.h>
#include <Box.h>
#include <Button.h>
#include <CheckBox.h>
#include <MediaDefs.h>
#include <MediaFormats.h>
#include <MenuField.h>
#include <MenuItem.h>
#include <OptionPopUp.h>
#include <PopUpMenu.h>
#include <Path.h>
#include <RadioButton.h>
#include <Screen.h>
#include <StringView.h>
#include <TextControl.h>
#include <Window.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>

static const int32 kAreaSelectionChanged = 'CaCh';
static const int32 MSG_FILE_FORMAT = 'FaDt';
static const int32 MSG_CODEC = 'Codc';
static const char *CODEC_DATA = "Codec";

TOutputView::TOutputView(BRect rect)
	:
	BView(rect, "Capture Options", B_FOLLOW_NONE, B_WILL_DRAW),
	fPlayingFPS(10)
{
	fCaptureArea.Set(0, 0, 0, 0);
	
	BRect outputBoxRect(10, 10, 220, 200);
	BBox *outputBox = new BBox(outputBoxRect, "Output");
	outputBox->SetLabel("Output");
	AddChild(outputBox);
				
	BRect optionRect(5, 15, 200, 40);
	const char *kTCLabel = "File name:"; 
	fFileName = new BTextControl(optionRect, "file name",
			kTCLabel, "/boot/home/outputVideo", new BMessage());
	
	outputBox->AddChild(fFileName);
	
	fFileName->SetDivider(StringWidth(kTCLabel) + 3);
	
	const char *kOutputMenuLabel = "Output Format:";
	optionRect.OffsetBy(0, 30);
	
	fOutputFileType = new BOptionPopUp(optionRect, "OutFormat",
			kOutputMenuLabel, new BMessage(MSG_FILE_FORMAT));
	outputBox->AddChild(fOutputFileType);
	
	fOutputFileType->MenuField()->SetDivider(StringWidth(kOutputMenuLabel));
		
	const char *kCodecMenuLabel = "Codec:";
	optionRect.OffsetBy(0, 30);
	BPopUpMenu *popUpMenu = new BPopUpMenu("Codecs");
	fCodecMenu = new BMenuField(optionRect, "OutCodec", kCodecMenuLabel, popUpMenu);
	outputBox->AddChild(fCodecMenu);
	
	fCodecMenu->SetDivider(StringWidth(kCodecMenuLabel) + 5);
	
	optionRect.OffsetBy(0, 40);
	
	BRect selectionRect;
	selectionRect.right = Bounds().right - 10;
	selectionRect.left = selectionRect.right - 210;
	selectionRect.top = Bounds().top + 10;
	selectionRect.bottom = selectionRect.top + 170;
	BBox *selectionBox = new BBox(selectionRect, "Capture Area");
	selectionBox->SetLabel("Capture Area");
	
	BRect radioRect(10, 25, 100, 40);
	fWholeScreen = new BRadioButton(radioRect, "screen frame", "Whole screen", new BMessage(kAreaSelectionChanged));
	selectionBox->AddChild(fWholeScreen);
	
	radioRect.OffsetBy(0, 30);
	fCustomArea = new BRadioButton(radioRect, "custom area", "Custom Area", new BMessage(kAreaSelectionChanged));
	selectionBox->AddChild(fCustomArea);
	
	BRect buttonRect = radioRect;
	buttonRect.OffsetBy(90, -5);
	fSelectArea = new BButton(buttonRect, "select area", "Select", new BMessage(kSelectArea));
	fSelectArea->SetEnabled(false);
	selectionBox->AddChild(fSelectArea);
	
	BRect viewRect = radioRect;
	viewRect.left += 25;
	viewRect.right = selectionBox->Bounds().right - 35;
	viewRect.bottom += 60;
	viewRect.OffsetBy(0, 30);
	fRectView = new TRectView(viewRect);
	selectionBox->AddChild(fRectView);
	
	fCaptureArea = BScreen(Window()).Frame();
	AddChild(selectionBox);
	
	SetAreaString(fCaptureArea);
	
	BRect checkBoxRect(selectionBox->Frame().left, selectionBox->Frame().bottom + 7, 
			430, selectionBox->Frame().bottom + 30);
	fMinimizeOnStart = new BCheckBox(checkBoxRect, "Minimize on start",
		"Minimize when recording starts", new BMessage());
	AddChild(fMinimizeOnStart);
	
	// fill in the list of available file formats
	media_file_format mff;
	media_format_family selected;
	int32 cookie = 0;
	bool firstFound = true;
	while (get_next_file_format(&cookie, &mff) == B_OK) {
		if (mff.capabilities & media_file_format::B_KNOWS_ENCODED_VIDEO) {
			fOutputFileType->AddOption(mff.pretty_name, mff.family);
			if (firstFound) {
				selected = mff.family;
				fOutputFileType->MenuField()->Menu()->ItemAt(0)->SetMarked(true);
				firstFound = false;
			}
		}	
	}
	BuildCodecMenu(selected);	
}


void
TOutputView::AttachedToWindow()
{
	BView::AttachedToWindow();
	SetViewColor(ui_color(B_PANEL_BACKGROUND_COLOR));
	
	fOutputFileType->SetTarget(this);
	fCustomArea->SetTarget(this);
	fWholeScreen->SetTarget(this);
	fWholeScreen->SetValue(B_CONTROL_ON);
}


void
TOutputView::MessageReceived(BMessage *message)
{
	switch (message->what) {
		case kAreaSelectionChanged:
		{
			fSelectArea->SetEnabled(fCustomArea->Value() == B_CONTROL_ON);
			BRect screenFrame = BScreen(Window()).Frame();
			if (fCaptureArea == screenFrame)
				break;
			if (fWholeScreen->Value() == B_CONTROL_ON)
				fCaptureArea = screenFrame;
						
			SetAreaString(fCaptureArea);		
			// Fall through
		}	
		case MSG_FILE_FORMAT:
		{
			BuildCodecMenu(FormatFamily());
			break;
		}
		case MSG_CODEC:
		{
			media_codec_info *info = NULL;
			ssize_t size;
			memset(&info, 0, sizeof(info));
			BMenuItem *item = fCodecMenu->Menu()->FindMarked();
			if (item != NULL)
				item->Message()->FindData(CODEC_DATA, B_SIMPLE_DATA, (const void**)&info, &size);
			if (info != NULL)
				fCodecInfo = *info;
			break;
		}
		default:
			BView::MessageReceived(message);
			break;
	}
}


void
TOutputView::SetPlayingFPS(uint32 fps)
{
	fPlayingFPS = fps;
}


bool
TOutputView::MinimizeOnStart() const
{
	return fMinimizeOnStart->Value() == B_CONTROL_ON;
}


void
TOutputView::RebuildCodecMenu()
{
	BuildCodecMenu(FormatFamily());
}


BPath
TOutputView::OutputFileName() const
{
	BPath path = fFileName->Text();
	return path;
}


bool
TOutputView::IncludeCursor() const
{
	return false;
}


void
TOutputView::SetCaptureArea(BRect rect)
{
	fCaptureArea.left = (int)rect.left;
	fCaptureArea.top = (int)rect.top;
	fCaptureArea.right = (int)rect.right;
	fCaptureArea.bottom = (int)rect.bottom;
	
	SetAreaString(fCaptureArea);
	BuildCodecMenu(FormatFamily());
}


BRect
TOutputView::CaptureArea() const
{
	return fCaptureArea;
}


media_format_family
TOutputView::FormatFamily() const
{
	return static_cast<media_format_family>(fOutputFileType->Value());
}


media_codec_info
TOutputView::Codec() const
{
	return fCodecInfo;
}


void
TOutputView::SetAreaString(BRect area)
{
	fRectView->SetRect(area);
	//fRectView->fCoordRect = area;
	//fRectView->Invalidate();
}


void 
TOutputView::BuildCodecMenu(media_format_family mf_family)
{
	memset(&fFormat, 0, sizeof(media_format));
	fFormat.type = B_MEDIA_RAW_VIDEO;
	fFormat.u.raw_video.display.line_width = uint32(fCaptureArea.Width()) + 1;
	fFormat.u.raw_video.display.line_count = uint32(fCaptureArea.Height() + 1);
	fFormat.u.raw_video.last_active = fFormat.u.raw_video.display.line_count - 1;
	
	// Since BScreen::BytesPerRow() is private, we have to take a screenshot
	// to get that information
	BBitmap *bitmap = NULL;
	BScreen screen;
	BRect smallRect(0, 0, 20, 20);
	screen.GetBitmap(&bitmap, false, &smallRect);
	fFormat.u.raw_video.display.bytes_per_row =
			(bitmap->BytesPerRow() /
			uint32(smallRect.Width() + 1)) * (fFormat.u.raw_video.display.line_width);
	
	delete bitmap;
	
	fFormat.u.raw_video.display.format = screen.ColorSpace();
	fFormat.u.raw_video.interlace = 1;	
	fFormat.u.raw_video.field_rate = fPlayingFPS; //Frames per second
	fFormat.u.raw_video.pixel_width_aspect = 1;		// square pixels
	fFormat.u.raw_video.pixel_height_aspect = 1;
	
	// suspend updates while we're rebuilding this menu in order to
	// reduce window flicker and other annoyances
	Window()->BeginViewTransaction();
	
	// find the full media_file_format corresponding to the given format family (e.g. AVI)
	status_t err;
	media_file_format mff;
	int32 cookie = 0;
	while ((err = get_next_file_format(&cookie, &mff)) == B_OK) {
		if (mff.family == mf_family)
			break;
	}

	// hmm, something is desperately wrong -- we couldn't find the "current" format family
	if (err) {
		fprintf(stderr, "ERROR:  couldn't find current media format family\n");
		exit(1);
	}

	BMenu *menu = fCodecMenu->Menu();
	if (menu != NULL)
		menu->RemoveItems(0, menu->CountItems(), true);
	
	int32 current = fCodecInfo.id;	// remember the current codec
	fCodecInfo.id = -2;						// magic invalid value

	// fill in the list of available codecs for this media format and file family
	media_codec_info codec;
	media_codec_info firstCodec;
	media_format outFormat;
	cookie = 0;
	bool foundCodec = false;
	while (get_next_encoder(&cookie, &mff, &fFormat, &outFormat, &codec) == B_OK) {
		BMessage *message = new BMessage(MSG_CODEC);
		message->AddData(CODEC_DATA, B_SIMPLE_DATA, &codec, sizeof(media_codec_info));
		BMenuItem *item = new BMenuItem(codec.pretty_name, message);
		menu->AddItem(item);
		if (!foundCodec) {
			foundCodec = true;
			firstCodec = codec;
		}
		if (codec.id == current) {
			fCodecInfo = codec;
			item->SetMarked(true);
		}
	}

	if (fCodecInfo.id == -2) {
		BMenuItem *item = menu->ItemAt(0);
		fCodecInfo = firstCodec;
		if (item)
			item->SetMarked(true);
	}
	// Make the app object the menu's message target
	menu->SetTargetForItems(this);
	
	// okay, we're done mucking with the menu
	Window()->EndViewTransaction();
}
